#include "code/game/grid/GameGrid.h"
#include "code/framework/Matrix.h"
#include "code/game/grid/MatrixContentObject.h"
#include <string>
#include "code/game/gamedef.h"
#include "code/game/FlagMatrix.h"
#include "code/game/pieces/impl/PieceSimple.h"
#include "code/game/pieces/impl/PieceShielded.h"
#include "code/game/pieces/abstract/AbstractPiece.h"

using namespace std;

GameGrid::GameGrid(int cols, int lines, int cellWidth, int cellHeight)
:Matrix<MatrixContentObject>(cols, lines)
{
	m_cellHeight=cellHeight;
	m_cellWidth=cellWidth;
	m_posX=int((480-getWidth())/2);
	m_posY=int((272-getHeight())/2);
	
	m_positions=new PositionsBuffer(cols, lines, m_posX, m_posY, cellWidth, cellHeight);
	fill();
}

GameGrid::~GameGrid()
{
	delete m_positions;
}

void GameGrid::fill()
{
	//debug("gamegrid fill");
	for(int i=0;i<getNbCols();i++)
	{
		for(int j=0;j<getNbLines();j++)
		{
			setAt(i, j, new MatrixContentObject(NULL));
		}
	}
}

GameGrid::GameGrid(int cols, int lines, int cellWidth, int cellHeight, int posX, int posY)
:Matrix<MatrixContentObject>(cols, lines)
{
	m_cellHeight=cellHeight;
	m_cellWidth=cellWidth;
	m_posX=posX;
	m_posY=posY;
	
	m_positions=new PositionsBuffer(cols, lines, m_posX, m_posY, cellWidth, cellHeight);
	
	fill();
}

void GameGrid::reset()
{
	for(int i=0;i<getNbCols();i++)
	{
		for(int j=0;j<getNbLines();j++)
		{
			MatrixContentObject *mco=getAt(i,j);
			mco->tryCleaning(true); //force
			if(mco->getMCOContent()!=NULL)
				mco->deleteMCOContent();
		}
	}
}

int GameGrid::getPosX()
{
	return m_posX;
}

int GameGrid::getPosY()
{
	return m_posY;
}

int GameGrid::getCellWidth()
{
	return m_cellWidth;
}

int GameGrid::getCellHeight()
{
	return m_cellHeight;
}

int GameGrid::getWidth()
{
	return m_cellWidth*m_cols;
}

int GameGrid::getHeight()
{
	return m_cellHeight*m_lines;
}

void GameGrid::update()
{
	for(int col=0;col<m_cols;col++)
	{
		for(int line=0;line<m_lines;line++)
		{
			//update de l'objet
			MatrixContentObject *mco=getAt(col, line);
			mco->update();
		}
	}	
}

Pos GameGrid::getContentPosition(int col, int line)
{
	return m_positions->getPixPos(col, line);
}

bool GameGrid::cellIsEmpty(int col, int line)
{
	//cas special ou la piece est en levitation au dessus de la grille:
	if(line<0 && col>=0 && col<m_cols)
	{
		return true;
	}
		
	if(col>=m_cols || line>=m_lines || col<0 || line<0)
	{
		return false;
	}
	
	//cas ou il y a un mco sans piece (normalement toujours)
	MatrixContentObject *mco=getAt(col, line);
	if(mco==NULL)
		fatalError("GameGrid::cellIsEmpty : une case n'a pas de mco! "+toString(col)+" "+toString(line));
	
	if(mco->isEmpty())
	{
		return true;
	}
	
	return false;
}

GridEventResult GameGrid::startExplosions(FlagMatrix *pack)
{
	GridEventResult ger;
	for(int i=0;i<pack->getNbCols();i++)
	{
		for(int j=0;j<pack->getNbLines();j++)
		{
			if(pack->isFlagSetAt(i, j))
			{
				ger.addExpl(startExplosions(Pos(i,j), pack));
			}
			//else: il ne s'agit pas d'une pièce qui vient d'atterrir
		}
	}
	return ger;
}

//TODO:optimiser ca, on peut éviter de passer 4 fois.
GridEventResult GameGrid::startExplosions(Pos p, FlagMatrix *pack)
{
	GridEventResult ger;
	
	//ETAPE0: si la piece est un joker, detruire toutes les pieces de la meme forme et/ou couleur
	//suivant ce que ca touche. si ca touche une piece de la meme couleur, toutes les pieces de
	//cette couleur explosent, si ca touche une piece de la meme forme toutes les pieces de cette
	//forme explosent, cumulable les 2
	//un joker ne peut pas venir d'en bas, il est toujours donné dans les paires
	if(getAt(p.x, p.y)->getPiece()->getType()==PIECE_TYPE_JOKER)
	{
		int shape=getAt(p.x, p.y)->getPiece()->getShape();
		int color=getAt(p.x, p.y)->getPiece()->getColor();
		
		int shapeToExplode=PIECE_SHAPE_NONE;
		int colorToExplode=PIECE_COLOR_NONE;
		
		//haut
		if(p.y-1>=0 && getAt(p.x, p.y-1)!=NULL && !getAt(p.x, p.y-1)->isEmpty())
		{
			if(pieceIsRightType(getAt(p.x, p.y-1)->getPiece()->getType(), true))
			{
				if(getAt(p.x, p.y-1)->getPiece()->getShape()==shape)
				{
					shapeToExplode=shape;
				}
				if(getAt(p.x, p.y-1)->getPiece()->getColor()==color)
				{
					colorToExplode=color;
				}
			}
		}
		
		//bas
		if(p.y+1<getNbLines() && getAt(p.x, p.y+1)!=NULL && !getAt(p.x, p.y+1)->isEmpty())
		{
		 	if(pieceIsRightType(getAt(p.x, p.y+1)->getPiece()->getType(), true))
			{
				if(getAt(p.x, p.y+1)->getPiece()->getShape()==shape)
				{
					shapeToExplode=shape;
				}
				if(getAt(p.x, p.y+1)->getPiece()->getColor()==color)
				{
					colorToExplode=color;
				}
			}
		}
		//gauche
		if(p.x-1>=0 && getAt(p.x-1, p.y)!=NULL && !getAt(p.x-1, p.y)->isEmpty())
		{
			if(pieceIsRightType(getAt(p.x-1, p.y)->getPiece()->getType(), true))
			{
				if(getAt(p.x-1, p.y)->getPiece()->getShape()==shape)
				{
					shapeToExplode=shape;
				}
				if(getAt(p.x-1, p.y)->getPiece()->getColor()==color)
				{
					colorToExplode=color;
				}
			}
		}
		
		//droite
		if(p.x+1<getNbCols() && getAt(p.x+1, p.y)!=NULL && !getAt(p.x+1, p.y)->isEmpty())
		{
			if(pieceIsRightType(getAt(p.x+1, p.y)->getPiece()->getType(), true))
			{
				if(getAt(p.x+1, p.y)->getPiece()->getShape()==shape)
				{
					shapeToExplode=shape;
				}
				if(getAt(p.x+1, p.y)->getPiece()->getColor()==color)
				{
					colorToExplode=color;
				}
			}
		}
		
		//et on explose tout!!
		if(shapeToExplode!=PIECE_SHAPE_NONE || colorToExplode!=PIECE_COLOR_NONE)
		{
			for(int i=0;i<getNbCols();i++)
			{
				for(int j=0;j<getNbLines();j++)
				{
					AbstractPiece *piece=getAt(i,j)->getPiece();
					if(piece!=NULL)
					{
						if(shapeToExplode!=PIECE_SHAPE_NONE && piece->getShape()==shapeToExplode)
						{
							getAt(i,j)->startExplosion(PIECE_MODE_SHAPE);
							if(!getAt(i,j)->isColorExploding())
								ger.addExplJoker();
						}
						if(colorToExplode!=PIECE_COLOR_NONE && piece->getColor()==colorToExplode)
						{
							getAt(i,j)->startExplosion(PIECE_MODE_COLOR);
							if(!getAt(i,j)->isShapeExploding())
								ger.addExplJoker();
						}
					}					
				}
			}
		}
		else
		{
			//faut être sûr d'exploser le joker quoi qu'il arrive
			getAt(p.x, p.y)->startExplosion(PIECE_MODE_COLOR);
			getAt(p.x, p.y)->startExplosion(PIECE_MODE_SHAPE);
			ger.addExplJoker();
		}
		
		return ger;
	}
	
	
	
	
	
	//ETAPE1: vérifier qu'un truc est pas en train d'exploser juste a coté
	//car dans ce cas, si ca correspond, elle explosera quoi qu'il arrive
	//idem si elle est elle même en train d'exploser, elle doit faire exploser
	//les alentours par contact. 
	//vrai pour les pieces normales et non shielded
	
	//haut
	if(p.y-1>=0 && getAt(p.x, p.y-1)!=NULL && !getAt(p.x, p.y-1)->isEmpty())
	{
		if(getAt(p.x, p.y)->getPiece()->getType()==PIECE_TYPE_NORMAL && getAt(p.x, p.y-1)->getPiece()->getType()==PIECE_TYPE_NORMAL)
		{
			//couleur
			if(	(getAt(p.x, p.y)->isColorExploding() || getAt(p.x, p.y-1)->isColorExploding()) && 
				getAt(p.x, p.y)->getPiece()->getColor()==getAt(p.x, p.y-1)->getPiece()->getColor())
			{
				FlagMatrix *flags=findExplosions(p, PIECE_MODE_COLOR, false);
				ger.addExplContact(explodeAll(flags, PIECE_MODE_COLOR));
				delete flags;
			}
			
			//forme
			if(	(getAt(p.x, p.y)->isShapeExploding() || getAt(p.x, p.y-1)->isShapeExploding()) && 
				getAt(p.x, p.y)->getPiece()->getShape()==getAt(p.x, p.y-1)->getPiece()->getShape())
			{
				FlagMatrix *flags=findExplosions(p, PIECE_MODE_SHAPE, false);
				ger.addExplContact(explodeAll(flags, PIECE_MODE_SHAPE));
				delete flags;
			}
		}
	}
			
	//bas
	if(p.y+1<getNbLines() && getAt(p.x, p.y+1)!=NULL && !getAt(p.x, p.y+1)->isEmpty())
	{
		if(getAt(p.x, p.y)->getPiece()->getType()==PIECE_TYPE_NORMAL && getAt(p.x, p.y+1)->getPiece()->getType()==PIECE_TYPE_NORMAL)
		{
			//couleur
			if(	(getAt(p.x, p.y)->isColorExploding() || getAt(p.x, p.y+1)->isColorExploding()) && 
				getAt(p.x, p.y)->getPiece()->getColor()==getAt(p.x, p.y+1)->getPiece()->getColor())
			{
				FlagMatrix *flags=findExplosions(p, PIECE_MODE_COLOR, false);
				ger.addExplContact(explodeAll(flags, PIECE_MODE_COLOR));
				delete flags;
			}
			
			//forme
			if(	(getAt(p.x, p.y)->isShapeExploding() || getAt(p.x, p.y+1)->isShapeExploding()) && 
				getAt(p.x, p.y)->getPiece()->getShape()==getAt(p.x, p.y+1)->getPiece()->getShape())
			{
				FlagMatrix *flags=findExplosions(p, PIECE_MODE_SHAPE, false);
				ger.addExplContact(explodeAll(flags, PIECE_MODE_SHAPE));
				delete flags;
			}
		}
	}
	
	//gauche
	if(p.x-1>=0 && getAt(p.x-1, p.y)!=NULL && !getAt(p.x-1, p.y)->isEmpty())
	{
		if(getAt(p.x, p.y)->getPiece()->getType()==PIECE_TYPE_NORMAL && getAt(p.x-1, p.y)->getPiece()->getType()==PIECE_TYPE_NORMAL)
		{
			//couleur
			if(	(getAt(p.x, p.y)->isColorExploding() || getAt(p.x-1, p.y)->isColorExploding()) && 
				getAt(p.x, p.y)->getPiece()->getColor()==getAt(p.x-1, p.y)->getPiece()->getColor())
			{
				FlagMatrix *flags=findExplosions(p, PIECE_MODE_COLOR, false);
				ger.addExplContact(explodeAll(flags, PIECE_MODE_COLOR));
				delete flags;
			}
			
			//forme
			if(	(getAt(p.x, p.y)->isShapeExploding() || getAt(p.x-1, p.y)->isShapeExploding()) && 
				getAt(p.x, p.y)->getPiece()->getShape()==getAt(p.x-1, p.y)->getPiece()->getShape())
			{
				FlagMatrix *flags=findExplosions(p, PIECE_MODE_SHAPE, false);
				ger.addExplContact(explodeAll(flags, PIECE_MODE_SHAPE));
				delete flags;
			}
		}
	}
	
	//droite
	if(p.x+1<getNbCols() && getAt(p.x+1, p.y)!=NULL && !getAt(p.x+1, p.y)->isEmpty())
	{
		if(getAt(p.x, p.y)->getPiece()->getType()==PIECE_TYPE_NORMAL && getAt(p.x+1, p.y)->getPiece()->getType()==PIECE_TYPE_NORMAL)
		{
			//couleur
			if(	(getAt(p.x, p.y)->isColorExploding() || getAt(p.x+1, p.y)->isColorExploding()) && 
				getAt(p.x, p.y)->getPiece()->getColor()==getAt(p.x+1, p.y)->getPiece()->getColor())
			{
				FlagMatrix *flags=findExplosions(p, PIECE_MODE_COLOR, false);
				ger.addExplContact(explodeAll(flags, PIECE_MODE_COLOR));
				delete flags;
			}
			
			//forme
			if(	(getAt(p.x, p.y)->isShapeExploding() || getAt(p.x+1, p.y)->isShapeExploding()) && 
				getAt(p.x, p.y)->getPiece()->getShape()==getAt(p.x+1, p.y)->getPiece()->getShape())
			{
				FlagMatrix *flags=findExplosions(p, PIECE_MODE_SHAPE, false);
				ger.addExplContact(explodeAll(flags, PIECE_MODE_SHAPE));
				delete flags;
			}
		}
	}
	
	
	//ETAPE2: rechercher si 4 (ou 5 pour shielded) pieces de meme couleur/formes se touchent
	//et qu'elles se trouvent en partie dans le pack et en partie sur le "sol"
	
	/////////////////////
	// pieces normales //
	/////////////////////
	
	
	//recherche par couleur, pieces normales
	FlagMatrix *simili=findExplosions(p, PIECE_MODE_COLOR, false); //va contenir des flags la ou des pieces sont similaires
	//on compte si y en a au moins 4
	int count=0;
	
	//au moins 1 dans le pack et un au sol
	bool inPack=false;
	bool inGround=false;
	
	for(int i=0;i<simili->getNbCols();i++)
	{
		for(int j=0;j<simili->getNbLines();j++)
		{
			if(simili->isFlagSetAt(i,j))
			{
				count++;
				if(pack!=NULL)
				{
					if(pack->isFlagSetAt(i,j)) //elle est dans le pack
					{
						inPack=true;
					}
					else //elle est dans le sol
					{
						inGround=true;
					}
				}
			}
		}
	}
	
	if(count>=GAMEDEF_NBPIECE_EXPLOSION && (pack==NULL || (inGround && inPack)))
	{
		//on démarre les explosions
		for(int i=0;i<simili->getNbCols();i++)
		{
			for(int j=0;j<simili->getNbLines();j++)
			{
				if(simili->isFlagSetAt(i,j))
				{
					if(!getAt(i,j)->isColorExploding())
					{
						ger.addExplNormal();
						getAt(i,j)->startExplosion(PIECE_MODE_COLOR);
					}
				}
			}
		}	
	}
	delete simili;
	
	
	//recherche par forme, pieces normales
	simili=findExplosions(p, PIECE_MODE_SHAPE, false);
	//on compte si y en a au moins 4
	count=0;
	//au moins 1 dans le pack et un au sol
	inPack=false;
	inGround=false;
	for(int i=0;i<simili->getNbCols();i++)
	{
		for(int j=0;j<simili->getNbLines();j++)
		{
			if(simili->isFlagSetAt(i,j))
			{
				count++;
				if(pack!=NULL)
				{
					if(pack->isFlagSetAt(i,j)) //elle est dans le pack
					{
						inPack=true;
					}
					else //elle est dans le sol
					{
						inGround=true;
					}
				}
			}
		}
	}
	if(count>=GAMEDEF_NBPIECE_EXPLOSION && (pack==NULL || (inGround && inPack)))
	{
		//on démarre les explosions
		for(int i=0;i<simili->getNbCols();i++)
		{
			for(int j=0;j<simili->getNbLines();j++)
			{
				if(simili->isFlagSetAt(i,j))
				{
					if(!getAt(i,j)->isShapeExploding())
					{
						ger.addExplNormal();
						getAt(i,j)->startExplosion(PIECE_MODE_SHAPE);
					}
				}
			}
		}	
	}
	delete simili;
	
	
	
	/////////////////////
	// pieces shielded //
	/////////////////////
	
	//recherche par couleur, pieces shielded
	simili=findExplosions(p, PIECE_MODE_COLOR, true); //va contenir des flags la ou des pieces sont similaires
	//on compte si y en a au moins 5
	count=0;
	
	//au moins 1 dans le pack et un au sol
	inPack=false;
	inGround=false;
	
	for(int i=0;i<simili->getNbCols();i++)
	{
		for(int j=0;j<simili->getNbLines();j++)
		{
			if(simili->isFlagSetAt(i,j))
			{
				count++;
				if(pack!=NULL)
				{
					if(pack->isFlagSetAt(i,j)) //elle est dans le pack
					{
						inPack=true;
					}
					else //elle est dans le sol
					{
						inGround=true;
					}
				}
			}
		}
	}
	
	if(count>=GAMEDEF_NBPIECE_EXPLOSION+1 && (pack==NULL || (inGround && inPack)))
	{
		//on démarre les explosions
		for(int i=0;i<simili->getNbCols();i++)
		{
			for(int j=0;j<simili->getNbLines();j++)
			{
				if(simili->isFlagSetAt(i,j))
				{
					if(!getAt(i,j)->isColorExploding())
					{
						ger.addExplNormal();
						getAt(i,j)->startExplosion(PIECE_MODE_COLOR);
					}
				}
			}
		}	
	}
	delete simili;
	
	
	//recherche par forme, pieces shielded
	simili=findExplosions(p, PIECE_MODE_SHAPE, true);
	//on compte si y en a au moins 5
	count=0;
	//au moins 1 dans le pack et un au sol
	inPack=false;
	inGround=false;
	for(int i=0;i<simili->getNbCols();i++)
	{
		for(int j=0;j<simili->getNbLines();j++)
		{
			if(simili->isFlagSetAt(i,j))
			{
				count++;
				if(pack!=NULL)
				{
					if(pack->isFlagSetAt(i,j)) //elle est dans le pack
					{
						inPack=true;
					}
					else //elle est dans le sol
					{
						inGround=true;
					}
				}
			}
		}
	}
	if(count>=GAMEDEF_NBPIECE_EXPLOSION+1 && (pack==NULL || (inGround && inPack)))
	{
		//on démarre les explosions
		for(int i=0;i<simili->getNbCols();i++)
		{
			for(int j=0;j<simili->getNbLines();j++)
			{
				if(simili->isFlagSetAt(i,j))
				{
					if(!getAt(i,j)->isShapeExploding())
					{
						ger.addExplNormal();
						getAt(i,j)->startExplosion(PIECE_MODE_SHAPE);
					}
				}
			}
		}	
	}
	delete simili;
	
	return ger;
}

int GameGrid::explodeAll(FlagMatrix *flags, int mode)
{
	int total=0;
	for(int i=0;i<flags->getNbCols();i++)
	{
		for(int j=0;j<flags->getNbLines();j++)
		{
			if(flags->isFlagSetAt(i,j))
			{
				if(!getAt(i,j)->isExploding(mode))
				{
					total++;
					getAt(i,j)->startExplosion(mode);
				}
			}
		}
	}	
	return total;
}

int GameGrid::explodeAll(int mode)
{
	int total=0;
	for(int i=0;i<getNbCols();i++)
	{
		for(int j=0;j<getNbLines();j++)
		{
			if(!getAt(i,j)->isExploding(mode) && getAt(i,j)->getPiece()!=NULL)
			{
				total++;
				getAt(i,j)->startExplosion(mode);
			}
		}
	}
	return total;
}

FlagMatrix *GameGrid::findExplosions(Pos p, int mode, bool withshielded, FlagMatrix *flags, int target)
{
	bool propagates=false;
	
	//positionnement du flag
	if(flags==NULL) //1er appel, on construit des flags à false
	{
		flags=new FlagMatrix(getNbCols(), getNbLines());
		if(getAt(p.x, p.y)->isEmpty() || getAt(p.x, p.y)->getPiece()->getType()==PIECE_TYPE_COMBO) 
		{
			//piece combo n'est pas un point de depart de combo, elle tombe et c'est tout
			return flags;
		}
		
		if(pieceIsRightType(getAt(p.x, p.y)->getPiece()->getType(), withshielded))
		{
			flags->setFlagAt(p.x, p.y); //la position de départ satisfait la condition, puisqu'on part de là
			target=getAt(p.x, p.y)->getPiece()->getByMode(mode);//c ce qu'on recherche, couleur ou forme
			propagates=true;
		}
		else
		{
			return flags;
		}
	}
	else //appels suivants
	{
//TODO: c ici le bug? isfalling!
		if(/*!getAt(p.x, p.y)->isFalling() &&*/ getAt(p.x, p.y)->getPiece()->getByMode(mode)==target 
			/*&& pieceIsRightType(getAt(p.x, p.y)->getPiece()->getType(), withshielded)*/) //TODO: test inutile
		{
			flags->setFlagAt(p.x, p.y);
			propagates=true;
		}
	}
	
	if(propagates)
	{
		//propagation
		//haut
		if(	p.y-1>=0 && getAt(p.x, p.y-1)!=NULL && !getAt(p.x, p.y-1)->isEmpty() && 
			!flags->isFlagSetAt(p.x, p.y-1) && pieceIsRightType(getAt(p.x, p.y-1)->getPiece()->getType(), withshielded))
		{
			/*flags=*/findExplosions(Pos(p.x, p.y-1), mode, withshielded, flags, target);
		}
		
		//bas
		if(	p.y+1<getNbLines() && getAt(p.x, p.y+1)!=NULL && !getAt(p.x, p.y+1)->isEmpty() && 
			!flags->isFlagSetAt(p.x, p.y+1) && pieceIsRightType(getAt(p.x, p.y+1)->getPiece()->getType(), withshielded))
		{
			/*flags=*/findExplosions(Pos(p.x, p.y+1), mode, withshielded, flags, target);
		}
		
		//gauche
		if(	p.x-1>=0 && getAt(p.x-1, p.y)!=NULL && !getAt(p.x-1, p.y)->isEmpty() && 
			!flags->isFlagSetAt(p.x-1, p.y) && pieceIsRightType(getAt(p.x-1, p.y)->getPiece()->getType(), withshielded))
		{
			/*flags=*/findExplosions(Pos(p.x-1, p.y), mode, withshielded, flags, target);
		}
		
		//droite
		if(	p.x+1<getNbCols() && getAt(p.x+1, p.y)!=NULL && !getAt(p.x+1, p.y)->isEmpty() && 
			!flags->isFlagSetAt(p.x+1, p.y) && pieceIsRightType(getAt(p.x+1, p.y)->getPiece()->getType(), withshielded))
		{
			/*flags=*/findExplosions(Pos(p.x+1, p.y), mode, withshielded, flags, target);
		}
	}
	
	return flags;
}

bool GameGrid::pieceIsRightType(int type, bool withshielded)
{
	if(type==PIECE_TYPE_NORMAL)
		return true;
	
	if(withshielded && type==PIECE_TYPE_SHIELDED)
		return true;
	
	return false;
}

void GameGrid::clean()
{
	for(int i=0;i<getNbCols();i++)
	{
		for(int j=0;j<getNbLines();j++)
		{
			MatrixContentObject *mco=getAt(i,j);
			if(mco==NULL)
				fatalError("SoloGame::cleanGrid() : mco is NULL");
			
			if(mco->tryCleaning()) //ca y est on a detruit la piece
			{
				//faut faire tomber la colonne
				for(int k=j-1;k>=0;k--)
				{
					MatrixContentObject *mcoup=getAt(i,k);
					if(mcoup==NULL)
						fatalError("SoloGame::cleanGrid() : mcoup is NULL");
					if(!mcoup->isEmpty())
						mcoup->startFalling();
				}
			}
			//else: rien a faire elle doit rester
		}
	}
}

int GameGrid::notifyCombo()
{
	int nb=0;
	for(int i=0;i<getNbCols();i++)
	{
		for(int j=getNbLines()-1;j>=0;j--)
		{
			MatrixContentObject *mco=getAt(i,j);
			if(mco==NULL)
				fatalError("SoloGame::notifyNewExplosions() : mco est NULL : "/*+toString(i)+","+toString(j)*/);
			
			if(mco->getPiece()!=NULL && mco->getPiece()->getType()==PIECE_TYPE_COMBO)
			{
				mco->getPiece()->specialEvent(PIECE_EVENT_COMBO_REDUCE);
				nb++;
			}
		}
	}
	return nb;
}

int GameGrid::explodeComboPieces()
{
	int res=0;
	for(int i=0;i<getNbCols();i++)
	{
		for(int j=getNbLines()-1;j>=0;j--)
		{
			MatrixContentObject *mco=getAt(i,j);
			if(mco==NULL)
				fatalError("SoloGame::explodeComboPieces() : mco est NULL : "/*+toString(i)+","+toString(j)*/);
			
			if(mco->getPiece()!=NULL && mco->getPiece()->getType()==PIECE_TYPE_COMBO)
			{
				if(mco->getPiece()->specialEvent(PIECE_EVENT_COMBO_EXPLODE))
				{
					res++;
					mco->setSpecialExploding();
				}
			}
		}
	}
	
	return res;
}

FlagMatrix *GameGrid::collapse()
{
	//construction de la matrice de flags indiquant les pieces qui viennent d'atterrir
	FlagMatrix *flags=new FlagMatrix(getNbCols(), getNbLines());
	
	for(int i=0;i<getNbCols();i++)
	{
		for(int j=getNbLines()-1;j>=0;j--)
		{			
			MatrixContentObject *mco=getAt(i,j);
			if(mco==NULL)
				fatalError("SoloGame::collapseGrid() : mco est NULL : "/*+toString(i)+","+toString(j)*/);
			
			if(mco->isFinishedFalling())
			{
				//il passe a la case du dessous
				if(j+1>=getNbLines())
					fatalError("SoloGame::collapseGrid() : une piece est tombee alors quelle etait en bas : "/*+toString(i)+","+toString(j)*/);
				
				MatrixContentObject *mcodown=getAt(i,j+1);
				if(mcodown==NULL)
					fatalError("SoloGame::collapseGrid() : mcodown est NULL");
				if(!mcodown->isEmpty())
					fatalError("SoloGame::collapseGrid() : une piece est tombee dans une case pleine : "/*+toString(i)+","+toString(j)*/);
				
				//c'est bien des set et non des replace			
				setAt(i,j+1,mco);
				setAt(i,j,mcodown/*new MatrixContentObject(NULL)*/);
				//delete mcodown;
				
				//est ce qu'elle doit continuer de tomber?
				if(j+2<getNbLines() && getAt(i,j+2)==NULL)
					fatalError("SoloGame::collapseGrid() : la case 2x en dessous de la piece est NULL");
				
				if(	j+2<getNbLines()
					&& (getAt(i,j+2)->isEmpty() || getAt(i,j+2)->isFalling()))
				{
					mco->restartFalling();
				}
				else
				{
					mco->dontFall();
					flags->setFlagAt(i,j+1);
				}
			}
		}
	}
	
	return flags;
}


FlagMatrix *GameGrid::cleanAndCollapse() //version optimisée 2 en 1
{
	//construction de la matrice de flags indiquant les pieces qui viennent d'atterrir
	FlagMatrix *flags=new FlagMatrix(getNbCols(), getNbLines());
	
	for(int i=0;i<getNbCols();i++)
	{
		for(int j=getNbLines()-1;j>=0;j--)
		{			
			MatrixContentObject *mco=getAt(i,j);
			
			//partie clean
			if(mco->tryCleaning()) //ca y est on a detruit la piece
			{
				//faut faire tomber la colonne
				for(int k=j-1;k>=0;k--)
				{
					MatrixContentObject *mcoup=getAt(i,k);
					if(mcoup==NULL)
						fatalError("SoloGame::cleanGrid() : mcoup is NULL");
					if(!mcoup->isEmpty())
						mcoup->startFalling();
				}
			}
			
			//partie collapse
			else if(mco->isFinishedFalling())
			{
				//il passe a la case du dessous
				if(j+1>=getNbLines())
					fatalError("SoloGame::collapseGrid() : une piece est tombee alors quelle etait en bas : "/*+toString(i)+","+toString(j)*/);
				
				MatrixContentObject *mcodown=getAt(i,j+1);
				if(mcodown==NULL)
					fatalError("SoloGame::collapseGrid() : mcodown est NULL");
				if(!mcodown->isEmpty())
					fatalError("SoloGame::collapseGrid() : une piece est tombee dans une case pleine : "/*+toString(i)+","+toString(j)*/);
				
				//c'est bien des set et non des replace			
				setAt(i,j+1,mco);
				setAt(i,j,mcodown/*new MatrixContentObject(NULL)*/);
				//delete mcodown;
				
				//est ce qu'elle doit continuer de tomber?
				if(j+2<getNbLines() && getAt(i,j+2)==NULL)
					fatalError("SoloGame::collapseGrid() : la case 2x en dessous de la piece est NULL");
				
				if(	j+2<getNbLines()
					&& (getAt(i,j+2)->isEmpty() || getAt(i,j+2)->isFalling()))
				{
					mco->restartFalling();
				}
				else
				{
					mco->dontFall();
					flags->setFlagAt(i,j+1);
				}
			}
		}
	}
	
	return flags;
}

bool GameGrid::isEvent()
{
	bool ret=false;
	
	for(int i=0;i<getNbCols();i++)
	{
		for(int j=getNbLines()-1;j>=0;j--)
		{	
			MatrixContentObject *mco=getAt(i,j);
			
			if(mco->isColorExploding() || mco->isShapeExploding() || mco->isSpecialExploding())
			{
				//TODO: voir si ca marche: phase de nettoyage suppl pour bug joker
				if(mco->getPiece()==NULL)
				{
					enableDebug();
					debug("nettoyage suppl!!");
					mco->setColorExploding(false);
					mco->setShapeExploding(false);
					mco->setSpecialExploding(false);
					debug("nettoyage ok");
					enableDebug(false);
				}
				else
				{
					ret=true;
				}
			}
			
			if(mco->isFalling())
			{
				ret=true;
			}
		}
	}
	return ret;
}
/*
bool GameGrid::addLine(list<AbstractPiece *> *pline) //false=gameover
{
	if(pline==NULL)
		fatalError("GameGrid::addLine : pline is NULL");
	
	bool notgameover=true;
	
	int column=0;
	list<AbstractPiece *>::iterator it;
	for (it=pline->begin(); it!=pline->end();it++)
	{
		//AbstractPiece *p_to_add=(*it)->clone();
		AbstractPiece *p=(*it);
		if(p!=NULL)
		{
			p=p->clone();
			MatrixContentObject *mco_to_add=new MatrixContentObject(p);
			
			//faut ajouter cette piece sous la colonne "column"
			for(int line=m_lines-1;line>=0;line--)
			{
				MatrixContentObject *mcotmp=getAt(column, line);
				if(mcotmp==NULL)
					fatalError("GameGrid::addLine : mco is NULL");
					
				//AbstractPiece * p_temp=mco->detachPiece();
				//mco->setPiece(p_to_add);
				setAt(column, line, mco_to_add);
				mco_to_add=mcotmp;
			}
			//fin des permutations vers le haut, si le dernier mco n'est pas vide, on a crevé le plafond, perdu.
			if(!mco_to_add->isEmpty())
			{
				if(mco_to_add->getPiece()->getType()!=PIECE_TYPE_NULL)
					notgameover=false;
			}
			delete mco_to_add; //sortie de la grille donc elle sera pas nettoyée.
		}
		
		column++;
	}
	
	return notgameover; //reussi, pas game over
}
*/

bool GameGrid::addLine(list<AbstractPiece *> *pline) //false=gameover
{
	if(pline==NULL)
		fatalError("GameGrid::addLine : pline is NULL");
	
	bool notgameover=true;
	
	int column=0;
	list<AbstractPiece *>::iterator it;
	for (it=pline->begin(); it!=pline->end();it++)
	{
		//AbstractPiece *p_to_add=(*it)->clone();
		AbstractPiece *p=(*it);
		if(p!=NULL)
		{
			p=p->clone();
			MatrixContentObject *mco_to_add=getAt(column, 0);
			if(!mco_to_add->isEmpty())
			{
				if(mco_to_add->getPiece()->getType()!=PIECE_TYPE_NULL)
					notgameover=false;	
			}
			mco_to_add->tryCleaning(true);
			if(mco_to_add->getMCOContent()!=NULL)
				mco_to_add->deleteMCOContent();
			mco_to_add->setPiece(p);
			
			//faut ajouter cette piece sous la colonne "column"
			for(int line=m_lines-1;line>=0;line--)
			{
				MatrixContentObject *mcotmp=getAt(column, line);
				
				setAt(column, line, mco_to_add);
				mco_to_add=mcotmp;
			}
		}
		
		column++;
	}
	
	return notgameover; //reussi, pas game over
}

void GameGrid::immediatePieceDestruction(int x, int y)
{
	//destruction
	MatrixContentObject *mco=getAt(x, y);
	mco->tryCleaning(true);
	
	//collapse de la colonne
	//faut faire tomber la colonne
	for(int k=y-1;k>=0;k--)
	{
		MatrixContentObject *mcoup=getAt(x,k);
		if(!mcoup->isEmpty())
			mcoup->startFalling();
	}
}















